# 25 "src/kernel_internals/parsing/logic_preprocess.mll"
 
  open Lexing
  type end_of_buffer = NEWLINE | SPACE | CHAR
  let preprocess_buffer = Buffer.create 1024

  let output_buffer = Buffer.create 1024
  (* Standard prohibits the predefined macros to be subject of a #define
     (or #undef) directive. We thus have to filter the definition of these
     macros from gcc's output (gcc emits a warning otherwise).
     The list of predefined macros is taken from C11 standard, in the order
     in which they are defined in Section 6.10.8
   *)
  let blacklisted_macros = [
    (* 6.10.8.1 mandatory macros. *)
    "__DATE__"; "__FILE"; "__LINE__"; "__STDC__"; "__STDC_HOSTED__";
    "__STDC_VERSION__"; "__TIME__";
    (* 6.10.8.2 environment macros *)
    "__STDC_ISO_10646__"; "__STDC_MB_MIGHT_NEQ_WC__";
    "__STDC_UTF_16__"; "__STDC_UTF_32__";
    (* 6.10.8.3 conditional feature macros *)
    "__STDC_ANALYZABLE__"; "__STDC_IEC_559__"; "__STDC_IEC_559_COMPLEX__";
    "__STDC_LIB_EXT1__"; "__STDC_NO_ATOMICS__"; "__STDC_NO_COMPLEX__";
    "__STDC_NO_THREADS__"; "__STDC_NO_VLA__";

    (* from TS 18661-1:2014 (for glibc >=2.35) *)
    "__STDC_IEC_60559_BFP__"; "__STDC_IEC_60559_COMPLEX__";

    (* expanding assert, an ACSL keyword, is not a good idea. *)
    "assert";
    (* __nonnull is predefined by Clang on macOS. *)
    "__nonnull";
  ]
  let is_newline = ref CHAR
  let curr_file = ref ""
  let curr_line = ref 1
  let has_annot = ref false

  let reset () =
    Buffer.clear preprocess_buffer;
    Buffer.clear output_buffer;
    is_newline := CHAR;
    curr_file := "";
    curr_line := 1;
    has_annot := false

  let backslash = "__ANNOT_BACKSLASH__"
  let annot_content = "__ANNOT_CONTENT__"
  let utf8_prefix = "__FC_UTF8_"

  let encode_utf8 c = utf8_prefix  ^ (string_of_int (Char.code c))

  let re_backslash = Str.regexp_string backslash
  let re_annot_content = Str.regexp_string annot_content
  let re_utf8 = Str.regexp (utf8_prefix ^ "\\([0-9]+\\)")

  let decode_utf8 s =
    let res = ref s in
    let start = ref 0 in
    try
      while true do
        let b = Str.search_forward re_utf8 !res !start in
        let e = Str.match_end () in
        let chr = Char.chr (int_of_string (Str.matched_group 1 !res)) in
        let buf = Bytes.of_string !res in
        Bytes.set buf b chr;
        Bytes.blit buf e buf (b+1) (String.length !res - e);
        res:= Bytes.sub_string buf 0 (String.length !res + 1 + b - e);
        start := b+1;
      done;
      assert false;
    with Not_found -> !res

  (* Delimiters for the various annotations in the preprocessing buffer.
     We have one delimiter for the beginning of an annotation (to discard
     #defines along the way), and three delimiters for the various ways
     an annotation can end:
      - on a normal line
      - with a newline
      - with a newline inside a comment (only for one-line annotations)
     When preprocessed annotations are inserted back in the main file, this will
     result in distinct translation to preserve line numbers while avoiding
     ill-formed annotations.
  *)
  let annot_beg =         "////////////////__ANNOT_BEG__"
  let annot_end =         "////////////////__ANNOT_END__"
  let annot_end_nl  =     "////////////////__ANNOT_END_NL__"
  let annot_end_comment = "////////////////__ANNOT_END_COMMENT__"

  let abort_preprocess reason =
    let source = {Cil_datatype.Position.unknown with Filepath.pos_path = Datatype.Filepath.of_string !curr_file;
                  pos_lnum = !curr_line;}
    in
    Kernel.error ~source
      "Can't preprocess annotation: %s\nSome annotations will be kept as is"
      reason

  let next_preprocessed file =
    let content = Buffer.create 80 in
    let rec ignore_content () =
      let s = input_line file in
      if not (Extlib.string_prefix annot_beg s) then ignore_content ()
    in
    let rec get_annot first =
      let s = input_line file in
      if Extlib.string_prefix annot_end s then false, Buffer.contents content
      else if Extlib.string_prefix annot_end_nl s then
        true, Buffer.contents content
      else if Extlib.string_prefix annot_end_comment s then begin
        Buffer.add_char content '\n';
        false, Buffer.contents content
      end else begin
        if not first then Buffer.add_char content '\n';
        Buffer.add_string content s;
        get_annot false
      end
    in
    let replace_backslash s = Str.global_replace re_backslash "\\\\" s in
    try
      ignore_content ();
      ignore (input_line file); (* ignore the #line directive *)
      let with_nl, content = get_annot true in
      with_nl, decode_utf8 @@ replace_backslash content
    with End_of_file ->
      Kernel.fatal
        "too few annotations in result file while pre-processing annotations"

  let output_result outfile preprocessed content =
    let rec aux = function
      | [] -> ()
      | [s] -> output_string outfile s
      | content :: rem ->
          output_string outfile content;
          output_string outfile "/*@";
          let with_nl, pp_content = next_preprocessed preprocessed in
          output_string outfile pp_content;
          output_string outfile "*/";
          if with_nl then output_char outfile '\n';
          aux rem
    in aux content

  let preprocess_annots suffix cpp outfile =
    if !has_annot then begin
      let debug =
        Kernel.debug_atleast 3 ||
          Kernel.is_debug_key_enabled Kernel.dkey_parser
      in
      let ppname =
        try Extlib.temp_file_cleanup_at_exit ~debug "ppannot" suffix
        with Extlib.Temp_file_error s ->
          Kernel.abort
            "Could not open temporary file for logic pre-processing: %s" s
      in
      let ppfile = open_out ppname in
      Buffer.output_buffer ppfile preprocess_buffer;
      close_out ppfile;
      let cppname = Extlib.temp_file_cleanup_at_exit ~debug "cppannot" suffix in
      let pp_cmd = cpp ppname cppname in
      Kernel.feedback ~dkey:Kernel.dkey_pp_logic
        "logic preprocessing with \"%s\"" pp_cmd;
      let res = Sys.command pp_cmd in
      let result_file =
        if res <> 0 then begin
          abort_preprocess "Preprocessor call exited with an error";
          if not debug then Extlib.safe_remove cppname;
          ppname
        end else cppname
      in
      let result = open_in result_file in
      let content =
        Str.split_delim re_annot_content (Buffer.contents output_buffer)
      in
      output_result outfile result content;
      close_in result
    end else begin
      Buffer.output_buffer outfile output_buffer
    end;
    flush outfile

  let add_preprocess_line_info () =
    Printf.bprintf
      preprocess_buffer "# %d %s \n" !curr_line !curr_file

  let make_newline () = incr curr_line

  let process_annot_start () =
    is_newline := CHAR;
    has_annot := true;
    Buffer.add_string output_buffer annot_content;
    Buffer.add_string preprocess_buffer annot_beg;
    Buffer.add_char preprocess_buffer '\n';
    add_preprocess_line_info()

# 195 "src/kernel_internals/parsing/logic_preprocess.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base =
   "\000\000\245\255\246\255\247\255\248\255\249\255\001\000\004\000\
    \017\000\000\000\031\000\000\000\000\000\000\000\001\000\001\000\
    \002\000\081\000\160\000\000\000\006\000\006\000\004\000\008\000\
    \018\001\028\000\010\000\030\000\031\000\254\255\029\000\032\000\
    \252\255\000\000\253\255\250\255\001\000\251\255\081\000\250\255\
    \251\255\252\255\253\255\003\000\001\000\255\255\254\255\036\000\
    \253\255\003\000\255\255\254\255\082\000\249\255\250\255\251\255\
    \252\255\085\000\253\255\254\255\255\255\212\000\249\255\250\255\
    \251\255\252\255\213\000\253\255\254\255\255\255\023\001\250\255\
    \252\255\253\255\083\000\255\255\251\255\254\255\024\001\250\255\
    \252\255\253\255\050\000\255\255\251\255\254\255\025\002\246\255\
    \247\255\248\255\249\255\250\255\251\255\252\255\028\000\254\255\
    \029\000\255\255\253\255\130\000\252\255\253\255\030\000\255\255\
    \254\255\153\001\250\255\251\255\215\000\254\255\255\255\252\255\
    \253\255\154\001\250\255\251\255\219\000\254\255\255\255\252\255\
    \253\255\155\001\252\255\253\255\254\255\255\255\156\001\250\255\
    \047\000\252\255\253\255\254\255\255\255\251\255\087\000\254\255\
    \255\255\088\000\254\255\255\255";
  Lexing.lex_backtrk =
   "\255\255\255\255\255\255\255\255\255\255\255\255\010\000\010\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\003\000\255\255\255\255\005\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\005\000\005\000\255\255\255\255\255\255\
    \255\255\002\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\006\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\006\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\005\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\005\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\009\000\255\255\
    \009\000\255\255\255\255\255\255\255\255\255\255\003\000\255\255\
    \255\255\255\255\255\255\255\255\005\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\005\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \005\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255";
  Lexing.lex_default =
   "\001\000\000\000\000\000\000\000\000\000\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\027\000\255\255\027\000\028\000\000\000\035\000\032\000\
    \000\000\255\255\000\000\000\000\255\255\000\000\039\000\000\000\
    \000\000\000\000\000\000\255\255\255\255\000\000\000\000\048\000\
    \000\000\255\255\000\000\000\000\053\000\000\000\000\000\000\000\
    \000\000\255\255\000\000\000\000\000\000\062\000\000\000\000\000\
    \000\000\000\000\255\255\000\000\000\000\000\000\071\000\000\000\
    \000\000\000\000\255\255\000\000\000\000\000\000\079\000\000\000\
    \000\000\000\000\255\255\000\000\000\000\000\000\088\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\255\255\000\000\
    \255\255\000\000\000\000\100\000\000\000\000\000\255\255\000\000\
    \000\000\106\000\000\000\000\000\255\255\000\000\000\000\000\000\
    \000\000\114\000\000\000\000\000\255\255\000\000\000\000\000\000\
    \000\000\122\000\000\000\000\000\000\000\000\000\127\000\000\000\
    \255\255\000\000\000\000\000\000\000\000\000\000\135\000\000\000\
    \000\000\138\000\000\000\000\000";
  Lexing.lex_trans =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\005\000\045\000\000\000\010\000\000\000\000\000\
    \000\000\000\000\000\000\026\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\026\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\007\000\010\000\000\000\000\000\002\000\
    \010\000\029\000\026\000\031\000\025\000\046\000\050\000\006\000\
    \030\000\026\000\051\000\025\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\255\255\010\000\
    \028\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\098\000\097\000\104\000\049\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\085\000\017\000\040\000\056\000\036\000\133\000\059\000\
    \033\000\136\000\139\000\000\000\020\000\013\000\014\000\017\000\
    \012\000\022\000\015\000\021\000\017\000\024\000\019\000\016\000\
    \009\000\017\000\023\000\042\000\054\000\077\000\000\000\060\000\
    \041\000\011\000\000\000\034\000\037\000\034\000\037\000\000\000\
    \043\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\009\000\103\000\000\000\084\000\000\000\
    \000\000\000\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\102\000\044\000\057\000\076\000\
    \018\000\058\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\000\000\000\000\000\000\000\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\000\000\000\000\000\000\000\000\065\000\068\000\
    \000\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\063\000\069\000\120\000\112\000\018\000\
    \004\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\024\000\255\255\255\255\255\255\255\255\
    \255\255\072\000\080\000\000\000\255\255\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \066\000\067\000\024\000\111\000\000\000\000\000\000\000\119\000\
    \000\000\073\000\000\000\000\000\000\000\000\000\000\000\081\000\
    \000\000\000\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\000\000\000\000\000\000\000\000\
    \000\000\255\255\055\000\000\000\000\000\000\000\000\000\136\000\
    \139\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\074\000\082\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\101\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\110\000\118\000\124\000\132\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\117\000\000\000\129\000\000\000\
    \109\000\000\000\000\000\130\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\125\000\128\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\064\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\108\000\116\000\000\000\
    \131\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\075\000\
    \083\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\095\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\092\000\087\000\089\000\087\000\087\000\087\000\087\000\
    \090\000\087\000\087\000\096\000\087\000\087\000\087\000\087\000\
    \094\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\093\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\091\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\107\000\115\000\123\000\132\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\255\255";
  Lexing.lex_check =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\044\000\255\255\007\000\255\255\255\255\
    \255\255\255\255\255\255\026\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\008\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\000\000\007\000\255\255\255\255\000\000\
    \010\000\028\000\026\000\006\000\026\000\043\000\047\000\000\000\
    \006\000\008\000\049\000\008\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\025\000\010\000\
    \027\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\094\000\096\000\102\000\047\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\082\000\017\000\038\000\052\000\030\000\128\000\057\000\
    \031\000\134\000\137\000\255\255\019\000\012\000\013\000\016\000\
    \007\000\009\000\014\000\020\000\021\000\023\000\011\000\015\000\
    \007\000\017\000\022\000\038\000\052\000\074\000\255\255\057\000\
    \038\000\007\000\255\255\033\000\036\000\033\000\036\000\255\255\
    \038\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\010\000\099\000\255\255\082\000\255\255\
    \255\255\255\255\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\099\000\038\000\052\000\074\000\
    \017\000\057\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\255\255\255\255\255\255\255\255\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\255\255\255\255\255\255\255\255\061\000\066\000\
    \255\255\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\061\000\066\000\116\000\108\000\018\000\
    \000\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\024\000\025\000\030\000\027\000\028\000\
    \031\000\070\000\078\000\255\255\047\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \061\000\066\000\024\000\108\000\255\255\255\255\255\255\116\000\
    \255\255\070\000\255\255\255\255\255\255\255\255\255\255\078\000\
    \255\255\255\255\024\000\024\000\024\000\024\000\024\000\024\000\
    \024\000\024\000\024\000\024\000\255\255\255\255\255\255\255\255\
    \255\255\038\000\052\000\255\255\255\255\255\255\255\255\134\000\
    \137\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\070\000\078\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\099\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\105\000\113\000\121\000\126\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\113\000\255\255\126\000\255\255\
    \105\000\255\255\255\255\126\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\121\000\126\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\061\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\105\000\113\000\255\255\
    \126\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\070\000\
    \078\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\105\000\113\000\121\000\126\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\086\000";
  Lexing.lex_base_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \010\000\000\000\036\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\001\000\000\000\000\000\000\000\000\000\000\000\002\000\
    \062\000\000\000\002\000\002\000\000\000\026\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000";
  Lexing.lex_backtrk_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\020\000\020\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000";
  Lexing.lex_default_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000";
  Lexing.lex_trans_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\001\000\017\000\009\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\009\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \001\000\017\000\009\000\001\000\023\000\000\000\000\000\000\000\
    \000\000\000\000\009\000\000\000\000\000\001\000\000\000\000\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\012\000\012\000\012\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\001\000\000\000\000\000\001\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\012\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\012\000\012\000\001\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\017\000\017\000\001\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\012\000\012\000\012\000\012\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_check_code =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\007\000\017\000\026\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\008\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \007\000\017\000\026\000\000\000\027\000\255\255\255\255\255\255\
    \255\255\255\255\008\000\255\255\255\255\010\000\255\255\255\255\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\010\000\255\255\255\255\024\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\024\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\016\000\021\000\023\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\024\000\024\000\
    \024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_code =
   "\255\004\255\255\006\255\005\255\255\006\255\255\005\255\006\255\
    \255\007\255\255\000\007\255\008\255\255\000\004\001\005\002\006\
    \003\008\255";
}

let rec main lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 9 (-1); __ocaml_lex_main_rec lexbuf 0
and __ocaml_lex_main_rec lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 221 "src/kernel_internals/parsing/logic_preprocess.mll"
                                                                      m
# 609 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 222 "src/kernel_internals/parsing/logic_preprocess.mll"
      (
        let blacklisted = List.mem m blacklisted_macros in
        if not blacklisted then
          Buffer.add_string preprocess_buffer (lexeme lexbuf);
        macro blacklisted lexbuf
      )
# 618 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
let
# 228 "src/kernel_internals/parsing/logic_preprocess.mll"
                                                       line
# 624 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 229 "src/kernel_internals/parsing/logic_preprocess.mll"
                                     file
# 629 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(2) lexbuf.Lexing.lex_mem.(3) in
# 230 "src/kernel_internals/parsing/logic_preprocess.mll"
    ( (try
        curr_line := (int_of_string line) -1
       with Failure _ -> curr_line:= -1);
      if file <> "" then curr_file := file;
      Buffer.add_string output_buffer (lexeme lexbuf);
      make_newline();
      main lexbuf
    )
# 640 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
let
# 238 "src/kernel_internals/parsing/logic_preprocess.mll"
                        c
# 646 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 3) in
# 238 "src/kernel_internals/parsing/logic_preprocess.mll"
                           ( (* Skip special doxygen comments. Use of '@'
                                instead of !Clexer.annot_char is intentional *)
        Buffer.add_string output_buffer (lexeme lexbuf);
        comment c lexbuf;)
# 653 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
let
# 242 "src/kernel_internals/parsing/logic_preprocess.mll"
                c
# 659 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 2) in
# 242 "src/kernel_internals/parsing/logic_preprocess.mll"
                   (
      if c = !Clexer.annot_char then begin
        process_annot_start ();
        annot lexbuf
      end else begin
        if c = '\n' then make_newline();
        Buffer.add_string output_buffer (lexeme lexbuf);
        comment c lexbuf;
      end)
# 671 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 4 ->
# 251 "src/kernel_internals/parsing/logic_preprocess.mll"
                      ( (* See comments for "/*@{" above *)
        Buffer.add_string output_buffer (lexeme lexbuf);
        oneline_comment lexbuf;
      )
# 679 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 5 ->
let
# 255 "src/kernel_internals/parsing/logic_preprocess.mll"
                c
# 685 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 2) in
# 255 "src/kernel_internals/parsing/logic_preprocess.mll"
                   (
      if c = !Clexer.annot_char then begin
        process_annot_start ();
        oneline_annot lexbuf
      end
      else if c = '\n' then begin
        make_newline ();
        Buffer.add_string output_buffer (lexeme lexbuf);
        main lexbuf
      end
      else begin
        Buffer.add_string output_buffer (lexeme lexbuf);
        oneline_comment lexbuf;
      end)
# 702 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 6 ->
# 269 "src/kernel_internals/parsing/logic_preprocess.mll"
         (
      make_newline (); Buffer.add_char output_buffer '\n'; main lexbuf )
# 708 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 7 ->
# 271 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( )
# 713 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 8 ->
# 272 "src/kernel_internals/parsing/logic_preprocess.mll"
        (
      Buffer.add_char output_buffer '"'; 
      c_string lexbuf )
# 720 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 9 ->
# 275 "src/kernel_internals/parsing/logic_preprocess.mll"
        (
      Buffer.add_char output_buffer '\'';
      c_char lexbuf )
# 727 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 10 ->
let
# 278 "src/kernel_internals/parsing/logic_preprocess.mll"
         c
# 733 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 278 "src/kernel_internals/parsing/logic_preprocess.mll"
           (
      Buffer.add_char output_buffer c;
      main lexbuf )
# 739 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_main_rec lexbuf __ocaml_lex_state

and macro blacklisted lexbuf =
   __ocaml_lex_macro_rec blacklisted lexbuf 38
and __ocaml_lex_macro_rec blacklisted lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 282 "src/kernel_internals/parsing/logic_preprocess.mll"
         (
      make_newline ();
      Buffer.add_char output_buffer '\n';
      macro blacklisted lexbuf
    )
# 755 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 289 "src/kernel_internals/parsing/logic_preprocess.mll"
       ( macro_comment blacklisted lexbuf )
# 760 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 290 "src/kernel_internals/parsing/logic_preprocess.mll"
      ( 
  if not blacklisted then
    Buffer.add_char preprocess_buffer '"';
  macro_string blacklisted lexbuf
)
# 769 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
# 295 "src/kernel_internals/parsing/logic_preprocess.mll"
      (
  if not blacklisted then
    Buffer.add_char preprocess_buffer '\'';
  macro_char blacklisted lexbuf
)
# 778 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 4 ->
# 300 "src/kernel_internals/parsing/logic_preprocess.mll"
       (
      if not blacklisted then
        Buffer.add_char preprocess_buffer '\n';
      make_newline ();
      Buffer.add_char output_buffer '\n';
      main lexbuf
    )
# 789 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 5 ->
let
# 307 "src/kernel_internals/parsing/logic_preprocess.mll"
       c
# 795 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 307 "src/kernel_internals/parsing/logic_preprocess.mll"
         (
           if not blacklisted then
             Buffer.add_char preprocess_buffer c;
           macro blacklisted lexbuf
         )
# 803 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_macro_rec blacklisted lexbuf __ocaml_lex_state

and macro_comment blacklisted lexbuf =
   __ocaml_lex_macro_comment_rec blacklisted lexbuf 47
and __ocaml_lex_macro_comment_rec blacklisted lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 313 "src/kernel_internals/parsing/logic_preprocess.mll"
       (
      make_newline ();

      macro_comment blacklisted lexbuf
    )
# 819 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 318 "src/kernel_internals/parsing/logic_preprocess.mll"
       ( macro blacklisted lexbuf )
# 824 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 319 "src/kernel_internals/parsing/logic_preprocess.mll"
     ( macro_comment blacklisted lexbuf )
# 829 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_macro_comment_rec blacklisted lexbuf __ocaml_lex_state

and macro_string blacklisted lexbuf =
   __ocaml_lex_macro_string_rec blacklisted lexbuf 52
and __ocaml_lex_macro_string_rec blacklisted lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 322 "src/kernel_internals/parsing/logic_preprocess.mll"
             s
# 842 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos (lexbuf.Lexing.lex_start_pos + 2) in
# 322 "src/kernel_internals/parsing/logic_preprocess.mll"
               (
  if not blacklisted then Buffer.add_string preprocess_buffer s;
  macro_string blacklisted lexbuf
  )
# 849 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 326 "src/kernel_internals/parsing/logic_preprocess.mll"
         (
  make_newline();
  Buffer.add_char output_buffer '\n';
  macro_string blacklisted lexbuf
)
# 858 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
let
# 331 "src/kernel_internals/parsing/logic_preprocess.mll"
            s
# 864 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos (lexbuf.Lexing.lex_start_pos + 2) in
# 331 "src/kernel_internals/parsing/logic_preprocess.mll"
              (
    if not blacklisted then Buffer.add_string preprocess_buffer s;
    macro_string blacklisted lexbuf
  )
# 871 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
# 335 "src/kernel_internals/parsing/logic_preprocess.mll"
       ( abort_preprocess "unterminated string in macro definition" )
# 876 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 4 ->
# 336 "src/kernel_internals/parsing/logic_preprocess.mll"
      ( abort_preprocess "unterminated string in macro definition" )
# 881 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 5 ->
# 337 "src/kernel_internals/parsing/logic_preprocess.mll"
      ( if not blacklisted then Buffer.add_char preprocess_buffer '"';
        macro blacklisted lexbuf )
# 887 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 6 ->
let
# 339 "src/kernel_internals/parsing/logic_preprocess.mll"
       c
# 893 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 339 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( if not blacklisted then Buffer.add_char preprocess_buffer c;
           macro_string blacklisted lexbuf )
# 898 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_macro_string_rec blacklisted lexbuf __ocaml_lex_state

and macro_char blacklisted lexbuf =
   __ocaml_lex_macro_char_rec blacklisted lexbuf 61
and __ocaml_lex_macro_char_rec blacklisted lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 342 "src/kernel_internals/parsing/logic_preprocess.mll"
            s
# 911 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos (lexbuf.Lexing.lex_start_pos + 2) in
# 342 "src/kernel_internals/parsing/logic_preprocess.mll"
              (
  if not blacklisted then Buffer.add_string preprocess_buffer s;
  macro_char blacklisted lexbuf
  )
# 918 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 346 "src/kernel_internals/parsing/logic_preprocess.mll"
         (
  make_newline();
  Buffer.add_char output_buffer '\n';
  macro_char blacklisted lexbuf
)
# 927 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
let
# 351 "src/kernel_internals/parsing/logic_preprocess.mll"
            s
# 933 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos (lexbuf.Lexing.lex_start_pos + 2) in
# 351 "src/kernel_internals/parsing/logic_preprocess.mll"
              (
    if not blacklisted then Buffer.add_string preprocess_buffer s;
    macro_char blacklisted lexbuf
  )
# 940 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
# 355 "src/kernel_internals/parsing/logic_preprocess.mll"
       ( abort_preprocess "unterminated char in macro definition" )
# 945 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 4 ->
# 356 "src/kernel_internals/parsing/logic_preprocess.mll"
      ( abort_preprocess "unterminated char in macro definition" )
# 950 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 5 ->
# 357 "src/kernel_internals/parsing/logic_preprocess.mll"
      ( if not blacklisted then Buffer.add_char preprocess_buffer '\'';
        macro blacklisted lexbuf )
# 956 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 6 ->
let
# 359 "src/kernel_internals/parsing/logic_preprocess.mll"
       c
# 962 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 359 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( if not blacklisted then Buffer.add_char preprocess_buffer c;
           macro_char blacklisted lexbuf )
# 967 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_macro_char_rec blacklisted lexbuf __ocaml_lex_state

and c_string lexbuf =
   __ocaml_lex_c_string_rec lexbuf 70
and __ocaml_lex_c_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 362 "src/kernel_internals/parsing/logic_preprocess.mll"
      ( abort_preprocess "unterminated string" )
# 979 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 363 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_string output_buffer (lexeme lexbuf); c_string lexbuf )
# 984 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 364 "src/kernel_internals/parsing/logic_preprocess.mll"
       ( Buffer.add_char output_buffer '"'; main lexbuf )
# 989 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
# 365 "src/kernel_internals/parsing/logic_preprocess.mll"
       ( make_newline ();
         Buffer.add_char output_buffer '\n';
         c_string lexbuf
       )
# 997 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 4 ->
# 369 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_string output_buffer (lexeme lexbuf); c_string lexbuf )
# 1002 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 5 ->
let
# 370 "src/kernel_internals/parsing/logic_preprocess.mll"
       c
# 1008 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 370 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_char output_buffer c; c_string lexbuf )
# 1012 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_c_string_rec lexbuf __ocaml_lex_state

and c_char lexbuf =
   __ocaml_lex_c_char_rec lexbuf 78
and __ocaml_lex_c_char_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 373 "src/kernel_internals/parsing/logic_preprocess.mll"
      ( abort_preprocess "unterminated char" )
# 1024 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 374 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_string output_buffer (lexeme lexbuf);
           c_char lexbuf )
# 1030 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 376 "src/kernel_internals/parsing/logic_preprocess.mll"
      ( Buffer.add_char output_buffer '\''; main lexbuf )
# 1035 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
# 377 "src/kernel_internals/parsing/logic_preprocess.mll"
       ( make_newline ();
         Buffer.add_char output_buffer '\n';
         c_char lexbuf
       )
# 1043 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 4 ->
# 381 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_string output_buffer (lexeme lexbuf); c_char lexbuf )
# 1048 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 5 ->
let
# 382 "src/kernel_internals/parsing/logic_preprocess.mll"
       c
# 1054 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 382 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_char output_buffer c; c_char lexbuf )
# 1058 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_c_char_rec lexbuf __ocaml_lex_state

and annot lexbuf =
   __ocaml_lex_annot_rec lexbuf 86
and __ocaml_lex_annot_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 385 "src/kernel_internals/parsing/logic_preprocess.mll"
          (
      if !is_newline = NEWLINE then
        Buffer.add_string preprocess_buffer annot_end_nl
      else begin
        Buffer.add_char preprocess_buffer '\n';
        Buffer.add_string preprocess_buffer annot_end;
      end;
      Buffer.add_char preprocess_buffer '\n';
      main lexbuf )
# 1078 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 394 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( is_newline := NEWLINE;
           incr curr_line;
           Buffer.add_char preprocess_buffer '\n';
           annot lexbuf )
# 1086 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 398 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_string preprocess_buffer "//";
           annot_comment lexbuf )
# 1092 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
# 400 "src/kernel_internals/parsing/logic_preprocess.mll"
        (
      if !is_newline = NEWLINE then is_newline:=SPACE;
      Buffer.add_char preprocess_buffer '@';
      annot lexbuf )
# 1100 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 4 ->
# 404 "src/kernel_internals/parsing/logic_preprocess.mll"
         (
      if !is_newline = NEWLINE then is_newline:=SPACE;
      Buffer.add_char preprocess_buffer ' ';
      annot lexbuf )
# 1108 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 5 ->
# 412 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( 
        is_newline := CHAR;
        Buffer.add_string preprocess_buffer backslash;
        annot lexbuf )
# 1116 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 6 ->
# 416 "src/kernel_internals/parsing/logic_preprocess.mll"
         (
        is_newline := CHAR;
        Buffer.add_char preprocess_buffer '\'';
        char annot lexbuf )
# 1124 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 7 ->
# 420 "src/kernel_internals/parsing/logic_preprocess.mll"
         (
        is_newline:=CHAR;
        Buffer.add_char preprocess_buffer '"';
        string annot lexbuf )
# 1132 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 8 ->
let
# 424 "src/kernel_internals/parsing/logic_preprocess.mll"
            c
# 1138 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 424 "src/kernel_internals/parsing/logic_preprocess.mll"
              (
     Buffer.add_string preprocess_buffer (encode_utf8 c);
     annot lexbuf
    )
# 1145 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 9 ->
let
# 428 "src/kernel_internals/parsing/logic_preprocess.mll"
         c
# 1151 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 428 "src/kernel_internals/parsing/logic_preprocess.mll"
           ( is_newline := CHAR;
             Buffer.add_char preprocess_buffer c;
             annot lexbuf )
# 1157 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_annot_rec lexbuf __ocaml_lex_state

and annot_comment lexbuf =
   __ocaml_lex_annot_comment_rec lexbuf 99
and __ocaml_lex_annot_comment_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 433 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( incr curr_line; is_newline:=NEWLINE;
           Buffer.add_char preprocess_buffer '\n';
           annot lexbuf
         )
# 1172 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 437 "src/kernel_internals/parsing/logic_preprocess.mll"
         (
        Buffer.add_char preprocess_buffer '\n';
        Buffer.add_string preprocess_buffer annot_end;
        Buffer.add_char preprocess_buffer '\n';
        main lexbuf )
# 1181 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 442 "src/kernel_internals/parsing/logic_preprocess.mll"
        ( abort_preprocess "eof in the middle of a comment" )
# 1186 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
let
# 443 "src/kernel_internals/parsing/logic_preprocess.mll"
         c
# 1192 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 443 "src/kernel_internals/parsing/logic_preprocess.mll"
           (
    Buffer.add_char preprocess_buffer c; annot_comment lexbuf )
# 1197 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_annot_comment_rec lexbuf __ocaml_lex_state

and char annot lexbuf =
   __ocaml_lex_char_rec annot lexbuf 105
and __ocaml_lex_char_rec annot lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 448 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( incr curr_line; is_newline:=NEWLINE;
           Buffer.add_char preprocess_buffer '\n';
           char annot lexbuf
         )
# 1212 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 452 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( is_newline:=CHAR;
           Buffer.add_char preprocess_buffer '\'';
           annot lexbuf )
# 1219 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 455 "src/kernel_internals/parsing/logic_preprocess.mll"
          ( is_newline:=CHAR;
            Buffer.add_string preprocess_buffer "\\'";
            char annot lexbuf )
# 1226 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
# 458 "src/kernel_internals/parsing/logic_preprocess.mll"
           ( is_newline:=CHAR;
            Buffer.add_string preprocess_buffer "\\\\";
            char annot lexbuf )
# 1233 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 4 ->
# 461 "src/kernel_internals/parsing/logic_preprocess.mll"
        ( abort_preprocess "eof while parsing a char literal" )
# 1238 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 5 ->
let
# 462 "src/kernel_internals/parsing/logic_preprocess.mll"
         c
# 1244 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 462 "src/kernel_internals/parsing/logic_preprocess.mll"
           ( is_newline:=CHAR;
             Buffer.add_char preprocess_buffer c;
             char annot lexbuf )
# 1250 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_char_rec annot lexbuf __ocaml_lex_state

and string annot lexbuf =
   __ocaml_lex_string_rec annot lexbuf 113
and __ocaml_lex_string_rec annot lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 467 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( incr curr_line; is_newline:=NEWLINE;
           Buffer.add_char preprocess_buffer '\n'; string annot lexbuf
         )
# 1264 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 470 "src/kernel_internals/parsing/logic_preprocess.mll"
        ( is_newline:=CHAR;
          Buffer.add_char preprocess_buffer '"'; annot lexbuf )
# 1270 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 472 "src/kernel_internals/parsing/logic_preprocess.mll"
           ( is_newline:=CHAR;
             Buffer.add_string preprocess_buffer "\\\"";
             string annot lexbuf )
# 1277 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
# 475 "src/kernel_internals/parsing/logic_preprocess.mll"
           ( is_newline:=CHAR;
             Buffer.add_string preprocess_buffer "\\\\";
             string annot lexbuf )
# 1284 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 4 ->
# 478 "src/kernel_internals/parsing/logic_preprocess.mll"
        ( abort_preprocess "eof while parsing a string literal" )
# 1289 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 5 ->
let
# 479 "src/kernel_internals/parsing/logic_preprocess.mll"
         c
# 1295 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 479 "src/kernel_internals/parsing/logic_preprocess.mll"
           ( is_newline:=CHAR;
             Buffer.add_char preprocess_buffer c;
             string annot lexbuf )
# 1301 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_string_rec annot lexbuf __ocaml_lex_state

and comment c lexbuf =
   __ocaml_lex_comment_rec c lexbuf 121
and __ocaml_lex_comment_rec c lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 485 "src/kernel_internals/parsing/logic_preprocess.mll"
        (
      Buffer.add_char output_buffer  '/';
      if c = '*' then
        main lexbuf
      else
        comment '/' lexbuf
      )
# 1319 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 492 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( make_newline (); Buffer.add_char output_buffer '\n';
           comment '\n' lexbuf )
# 1325 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 494 "src/kernel_internals/parsing/logic_preprocess.mll"
        ( abort_preprocess "eof while parsing C comment" )
# 1330 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
let
# 495 "src/kernel_internals/parsing/logic_preprocess.mll"
         c
# 1336 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 495 "src/kernel_internals/parsing/logic_preprocess.mll"
           ( Buffer.add_char output_buffer c; comment c lexbuf )
# 1340 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_comment_rec c lexbuf __ocaml_lex_state

and oneline_annot lexbuf =
   __ocaml_lex_oneline_annot_rec lexbuf 126
and __ocaml_lex_oneline_annot_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 498 "src/kernel_internals/parsing/logic_preprocess.mll"
             (
      incr curr_line;
      Buffer.add_char preprocess_buffer '\n';
      Buffer.add_string preprocess_buffer annot_end_nl;
      Buffer.add_char preprocess_buffer '\n';
      main lexbuf )
# 1357 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 504 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_string preprocess_buffer backslash;
           oneline_annot lexbuf )
# 1363 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 506 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_char preprocess_buffer '\'';
           char oneline_annot lexbuf )
# 1369 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
# 508 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_char preprocess_buffer '"';
           string oneline_annot lexbuf )
# 1375 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 4 ->
# 510 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_string preprocess_buffer "//";
           oneline_annot_comment lexbuf )
# 1381 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 5 ->
let
# 512 "src/kernel_internals/parsing/logic_preprocess.mll"
         c
# 1387 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 512 "src/kernel_internals/parsing/logic_preprocess.mll"
           ( Buffer.add_char preprocess_buffer c;
             oneline_annot lexbuf )
# 1392 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_oneline_annot_rec lexbuf __ocaml_lex_state

and oneline_annot_comment lexbuf =
   __ocaml_lex_oneline_annot_comment_rec lexbuf 134
and __ocaml_lex_oneline_annot_comment_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 516 "src/kernel_internals/parsing/logic_preprocess.mll"
             (
       incr curr_line;
       Buffer.add_char preprocess_buffer '\n';
       Buffer.add_string preprocess_buffer annot_end_comment;
       Buffer.add_char preprocess_buffer '\n';
       main lexbuf )
# 1409 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
let
# 522 "src/kernel_internals/parsing/logic_preprocess.mll"
         c
# 1415 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 522 "src/kernel_internals/parsing/logic_preprocess.mll"
           ( Buffer.add_char preprocess_buffer c;
             oneline_annot_comment lexbuf )
# 1420 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_oneline_annot_comment_rec lexbuf __ocaml_lex_state

and oneline_comment lexbuf =
   __ocaml_lex_oneline_comment_rec lexbuf 137
and __ocaml_lex_oneline_comment_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 528 "src/kernel_internals/parsing/logic_preprocess.mll"
      ( make_newline();
        Buffer.add_string output_buffer (lexeme lexbuf);
        main lexbuf)
# 1434 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
let
# 531 "src/kernel_internals/parsing/logic_preprocess.mll"
         c
# 1440 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 531 "src/kernel_internals/parsing/logic_preprocess.mll"
           ( Buffer.add_char output_buffer c;
             oneline_comment lexbuf)
# 1445 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_oneline_comment_rec lexbuf __ocaml_lex_state

;;

# 534 "src/kernel_internals/parsing/logic_preprocess.mll"
 
  let file suffix cpp filename =
    reset ();
    let debug = Kernel.is_debug_key_enabled Kernel.dkey_parser in
    let scan_references = Kernel.EagerLoadSources.get () in
    match Parse_env.open_source ~scan_references filename with
    | Error msg -> Kernel.abort "logic_preprocess: %s" msg
    | Ok source ->
      let lex = Lexing.from_string source in
      let ppname =
        Extlib.temp_file_cleanup_at_exit ~debug
          (Filename.basename filename) ".pp"
      in
      let fp_of_string = Filepath.Normalized.of_string in
      let workdir_opt = Parse_env.get_workdir (fp_of_string filename) in
      Option.iter
        (fun workdir -> Parse_env.set_workdir (fp_of_string ppname) workdir)
        workdir_opt;
      let ppfile = open_out ppname in
      main lex;
      preprocess_annots suffix cpp ppfile;
      close_out ppfile;
      Datatype.Filepath.of_string ppname

# 1477 "src/kernel_internals/parsing/logic_preprocess.ml"
